
#include <stdio.h>
#include <dos.h>
#include <bios.h>
#include <signal.h>
#include <ctype.h>
#include <string.h>
#ifdef __BORLANDC__
#include <conio.h>
#endif

#define cprintf     printf /**/

#define KEY_SPECIAL 0x8000
#define KEY_PAD     0x4000
#define KEY_OTHER   0x2000
#define KEY_DIRECT  0x1000
#define KEY_ALT     0x0800
#define KEY_ALTGR   0x0400
#define KEY_CTRL    0x0200
#define KEY_SHIFT   0x0100

/* Internal use only */
#define KEY_CTRL2   0x0002
#define KEY_SHIFT2  0x0001

#define KEYTEST
int last_bios_key = -1;
int last_ctrl_flg = -1;
int last_stuffed  = -1;
int bios_stuff = 0;
int pending_keyscan = 0;	/* A scancode for which we expect a bios key */

#define KEYBD       0x09
#define STATE_FLAGS (*(int  far *)0x00400017L)	/* BIOS Shift state */
#define STATE_ALTCD (*(char far *)0x00400019L)	/* BIOS ALT-Compose byte */

int get_key(void);
void ( interrupt far *orig_keybd)();
void interrupt far keybd();
static int kbd_type = -1;

int scrcol = 0;
int keymode = 1;

void
main(int argc, char ** argv)
{
   int ch, by;
   union REGS regs;
   int i;
   int running = -1;

   for(i=1; i<argc; i++) if( argv[i][0] == '-' || argv[i][0] == '/' )
   switch(argv[i][1])
   {
   case 'x': case 'X': keymode |= 4; break;
   case 'u': case 'U': keymode &= ~3; break;
   }
   else if( pending_keyscan )
        bios_stuff      = strtol(argv[i], (void*)0, 0);
   else 
   {
        pending_keyscan = strtol(argv[i], (void*)0, 0);
        running = 20;
   }

   if( keymode & 4 ) { keymode &= ~4; kbd_type = 0 ; }

   signal(SIGINT, SIG_IGN);

   cprintf("Keyboard testing: Press Escape to exit\r\n");
   orig_keybd = _dos_getvect(KEYBD);
   _dos_setvect(KEYBD, keybd);

   do
   {
      ch = get_key();
      if( (ch &0x80FF) == '\033' ) running = ((unsigned)running>>2);
      if( running > 0 ) running--;

      if( last_bios_key != -1 ||
	  last_stuffed  != -1 ||
	  last_ctrl_flg != -1 )
      {
         if( scrcol ) cprintf(">\r\n"); scrcol=0;
         if( last_bios_key != -1 ) cprintf( "BIOS %04x ", last_bios_key);
         else                      cprintf( "BIOS ---- ");
         if( last_stuffed != -1 )  cprintf( "STUFFED %04x ", last_stuffed);
         else                      cprintf( "STUFFED ---- ");
         if( last_ctrl_flg != -1 ) cprintf( "FLAGS %04x ", last_ctrl_flg);
         else                      cprintf( "FLAGS ---- ");
         cprintf( "\r\n");

      }

      if( ch == -1 )
         continue;

      if( (ch&0xFF00) == KEY_SPECIAL + KEY_DIRECT + KEY_OTHER )
      /* It's a scan code copied in */
      {
	 if( scrcol > 72 ) { cprintf(">\r\n"); scrcol = 0; }
         if( scrcol == 0 ) { cprintf("<"); scrcol ++; }

	 if( (ch&0xFF) != 0xFA ) /* Not interested in Acks */
	    write_scan(ch&0xFF);
         if( STATE_FLAGS & 0x400 )
	 {
	    STATE_FLAGS &= ~0x400;
	    cprintf("Y"); scrcol++;
	 }
         continue;
      }

      if( scrcol ) cprintf(">\r\n"); scrcol=0;

      cprintf("Got key = 0x%04x is: ", ch);

      if( ch & KEY_SPECIAL )
      {
	 cprintf("Special-");
	 if( ch & KEY_CTRL )    cprintf("Ctrl-");
	 if( ch & KEY_SHIFT )   cprintf("Shift-");
	 if( ch & KEY_ALT )     cprintf("Alt-");
	 if( ch & KEY_ALTGR )   cprintf("AltGr-");
	 if( ch & KEY_DIRECT )  cprintf("Scan-");
	 by = (ch & 0xFF);
      }
      else 
      {
	 cprintf("ASCII-");
	 by = (ch & 0xFFF);
	 if( ch & KEY_DIRECT )  cprintf("Code-");
      }
      if( ch & KEY_PAD )     cprintf("KeyPad-");
      if( ch & KEY_OTHER )   cprintf("Other-");

      switch( ch & KEY_SPECIAL )
      {
      case KEY_SPECIAL:
	 if( ch & KEY_DIRECT )
	    cprintf("0x%02x\r\n", by);
	 else if((by&0xF0)==0x80 )
	    cprintf("F%d\r\n", by&0xF);
	 else if((by&0xF0)==0x90 )
	    cprintf("K%d\r\n", by&0xF);
	 else
	 {
      case 0:
	    if( by >= ' ' && by <= '~' && by != '\\' && by != '\'' )
	       cprintf("'%c'\r\n", by);
	    else
	       cprintf("0x%02x\r\n", by);
	 }
      }

      scrcol=4;
      cprintf(": ");
      if( keymode ) write_doscode(ch); else write_xenix(ch);
      cprintf(" <");
   }
   while( running );

   _dos_setvect(KEYBD, orig_keybd);

   while( get_key() != -1 ) ;

   if( scrcol ) { cprintf(">\r\n"); scrcol = 0; }

   exit(0);
}

void xmit(int ch)
{
   ch &= 0xFF;
   if( ch < ' ' )
   {
      cprintf("^%c", ch+'@');
      scrcol++;
   }
   else
      cprintf("%c", ch);
   scrcol++;
}

/****************************************************************************/

unsigned char scan_desc[] = {
 0,   0,   '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '=', 8,   0,  
 'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', 10,  'C', 'a', 's',
 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', 0,   0,   'S', 0,   'z', 'x', 'c', 'v',
 'b', 'n', 'm', ',', '.', '/', 'T', 0,   'A', ' ', 0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 'E', 'F', 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  
 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0
};

char hextbl[] = "0123456789abcdef";

/* Write scancodes, in this mode you really need a compatible program on
 * the host, but you can just about type 'reset<cr>' and have it work
 * if the "xmit('_')" is changed to an "xmit('~'); xmit(8);"
 */
write_scan(scan)
int scan;
{
static int lastscan = 0x100;
   int by;

   scan &= 0xFF;

   if( scan == lastscan+0x80 )
      xmit('_');
   else if( by = scan_desc[scan] )
      xmit(by);
   else if( by = scan_desc[scan&0x7F] )
   {
      xmit('!');
      xmit(by);
   }
   else
      { xmit('~'); xmit(hextbl[scan>>4]); xmit(hextbl[scan&0xF]); }
   lastscan = scan;
}

/*
 * This keyboard gives a pattern of control sequences that makes just about
 * every keypress distinct. It does need a compatible host program but its
 * useable as a generic terminal type unlike the scancodes.
 */
write_doscode(ch)
int ch;
{
#define CT_KEY 0x1E
   int by = (ch & 0xFF);

   /* First we swap round ^M and Return, likewise BS, Tab and Esc  */
   if( ( ch & 0xFF00 ) == KEY_OTHER || ( ch & 0xFF00 ) == 0 )
   {
      /* Make it so we can detect the 'other' keys */
      if( strchr("\b\t\r\n\033\003\177", by) )
      {
	 if( ch & KEY_OTHER )
	    ch ^= KEY_OTHER;
	 else
	    by = ((ch = KEY_SPECIAL+KEY_CTRL+tolower(by+'@')) & 0xFF);
      }
   }

   if( ch & 0xFF00 )
   {
      if( ch & KEY_SPECIAL )
      {
static char xkeys[] = "0123456789dnhspc";
static char fkeys[] = "01234567890-=ABC";
         if( by == 0 ) return;

	 if( ch & KEY_ALT )     { xmit(CT_KEY); xmit('a'); }
	 if( ch & KEY_CTRL )    { xmit(CT_KEY); xmit('c'); }
	 if( ch & KEY_SHIFT )   { xmit(CT_KEY); xmit('s'); }
	 if( ch & KEY_ALTGR )   { xmit(CT_KEY); xmit('g'); }
	 if( ch & KEY_PAD )     { xmit(CT_KEY); xmit('k'); }
	 /* if( ch & KEY_OTHER )   { xmit(CT_KEY); xmit('o'); } */

	 if( ch & KEY_DIRECT )
	 {
	    xmit(CT_KEY); xmit('~');
            xmit(hextbl[by>>4]); xmit(hextbl[by&0xF]);
	 }
         else if((by&0xF0)==0x80 )
	 {
	    if( by == 0x80 )
            {
	       xmit('\033');
	       xmit('\033');
            }
	    else
	    {
	       xmit(CT_KEY);
	       xmit(fkeys[by&0xF]);
	    }
	 }
	 else if((by&0xF0)==0x90 )
	 {
	    xmit(CT_KEY);
	    xmit('K');
	    switch(by)
	    {
	    case 0x9D: xmit('s'+ ((STATE_FLAGS&0x10)?'A'-'a':0)); break;
	    case 0x9B: xmit('n'+ ((STATE_FLAGS&0x20)?'A'-'a':0)); break;
	    case 0x9F: xmit('c'+ ((STATE_FLAGS&0x40)?'A'-'a':0)); break;
	    default: xmit(xkeys[by&0xF]); break;
	    }
         }
	 else
	    xmit(by);
      }
      else
      {
	 if( ch & KEY_DIRECT )
	 {
	    char buf[4], *p;
	    sprintf(buf, "%d", by);
	    xmit(CT_KEY); xmit('A');
	    for(p=buf; *p; p++)
	    {
	       xmit(CT_KEY); xmit('k'); xmit(*p);
	    }
	    xmit(CT_KEY); xmit(' ');
	 }
	 else
	 {
	    if( ch & KEY_PAD )     { xmit(CT_KEY); xmit('k'); }
	    if( ch & KEY_OTHER )   { xmit(CT_KEY); xmit('o'); }
	    xmit(by);
	 }
      }
   }
   else if( by == CT_KEY )
   {
      xmit(CT_KEY); xmit('c'); xmit(tolower(CT_KEY+'@'));
   }
   else
   {
      xmit(by);
      if( by == '\033' )
         xmit(by);
   }
}

/*
 * This generates key sequences compatible with the SCO-Xenix console.
 * In addition many extra key combinations generate distinct codes.
 * But there are normally several ways of generating each standard key.
 *
 * This key mapping is designed for easy use with terminfo nevertheless
 * it has potentially 96 function key + 104 alt/ctrl/shift&letter codes.
 */
write_xenix(ch)
int ch;
{
   int by;
                    /*  012345678901234567890123456   789012 */
static char normal[] = " MNOPQRSTUVWX   LFBGDECHAI\177      ";
static char shift [] = " YZabcdefghij   0123456789\177      ";
static char ctrl  [] = " klmnopqrstuv             \177      ";
static char ctrshf[] = " wxyz@[\\]^_`{             \177      ";

   /* This seems to be a pain if it stays */
   if( ch == KEY_SPECIAL+KEY_SHIFT+'\r' ) ch = '\r';
   if( ch == KEY_SPECIAL+KEY_SHIFT+' '  ) ch = ' ';

   if( ch & KEY_SPECIAL )
   {
      if( ch & (KEY_ALT|KEY_ALTGR|KEY_DIRECT) )
         write_doscode(ch);
      else if( (ch & 0xE0) != 0x80 )
         write_doscode(ch);
      else
      {
         char * keys;
	 if( ch&KEY_SHIFT )
	 {
	    if( ch&KEY_CTRL ) keys = ctrshf; else keys = shift;
	 }
	 else
	 {
	    if( ch&KEY_CTRL ) keys = ctrl; else keys = normal;
	 }
         by = keys[ch&0x1F];
	 if( by == ' ' ) write_doscode(ch);
	 else if( by == '\177' || by >= '0' && by <= '9' )
	    xmit(by);
         else
	 {
	    xmit('\033');
	    xmit('[');
	    xmit(by);
	 }
      }
   }
   else
      xmit(ch&0xFF);
}

/****************************************************************************/

void interrupt far keybd()
{
#ifdef __BORLANDC__
register int i;

   /* Save scancodes ... */
   if( kbd_type > 0 )
   {
      i = (inp(0x60) & 0xFF);
      _CX = 0xFF00 + i;
      _AH = 0x05; geninterrupt(0x16);

      /* Disable C-A-D, C-A-Esc and Co */
      if(i==0x53 || i==1 || i==0x1C || i == 0x4A || i == 0x4E )
      {
         if( ( (i=STATE_FLAGS) & 0xC ) == 0xC )
         {
	    STATE_FLAGS = (i & ~0x030C);	/* Turn off Ctrl-Alt */
	    (*orig_keybd)();
	    STATE_FLAGS = i;
	    return;
         }
      }
      /* Disable Pause */
      else if( i == 0xE1 )
         STATE_FLAGS &= ~0x0800;
   }
#endif

#if 0
for(i=0; i<2500; i++)
   if( STATE_FLAGS == 0 ) break;
#endif

   _chain_intr(orig_keybd);
}

/****************************************************************************/

static int bioskeydecode[] = {
/*00*/	0x0000, 0x3280, 0x3131, 0x3132, 0x3133, 0x3134, 0x3135, 0x3136,
	0x3137, 0x3138, 0x3139, 0x3130, 0x312d, 0x313d, 0x3308, 0x1209,
/*10*/	0x3171, 0x3177, 0x3165, 0x3172, 0x3174, 0x3179, 0x3175, 0x3169,
	0x316f, 0x3170, 0x315b, 0x315d, 0x330d, 0x0000, 0x3161, 0x3173,
/*20*/	0x3164, 0x3166, 0x3167, 0x3168, 0x316a, 0x316b, 0x316c, 0x313b,
	0x0127, 0x0160, 0x0000, 0x045c, 0x317a, 0x3178, 0x3163, 0x3176,
/*30*/	0x3162, 0x316e, 0x316d, 0x312c, 0x312e, 0x312f, 0x0000, 0x0700,
	0x0000, 0x3520, 0x0000, 0x0481, 0x0482, 0x0483, 0x0484, 0x0485,
/*40*/	0x0486, 0x0487, 0x0488, 0x0489, 0x048a, 0x0600, 0x0000, 0x0697,
	0x0698, 0x0699, 0x072d, 0x0694, 0x0695, 0x0696, 0x072b, 0x0691,
/*50*/	0x0692, 0x0693, 0x0690, 0x069a, 0x1481, 0x1482, 0x1483, 0x1484,
	0x1485, 0x1486, 0x1487, 0x1488, 0x1489, 0x148a, 0x2481, 0x2482,
/*60*/	0x2483, 0x2484, 0x2485, 0x2486, 0x2487, 0x2488, 0x2489, 0x248a,
	0x3481, 0x3482, 0x3483, 0x3484, 0x3485, 0x3486, 0x3487, 0x3488,
/*70*/	0x3489, 0x348a, 0x0000, 0x2694, 0x2696, 0x2691, 0x2693, 0x2697,
	0x3131, 0x3132, 0x3133, 0x3134, 0x3135, 0x3136, 0x3137, 0x3138,
/*80*/	0x3139, 0x3130, 0x002d, 0x003d, 0x2699, 0x048b, 0x048c, 0x148b,
	0x148c, 0x248b, 0x248c, 0x348b, 0x348c, 0x2698, 0x272d, 0x2695,
/*90*/	0x272b, 0x2692, 0x2690, 0x269a, 0x2009, 0x272f, 0x272a, 0x3697,
	0x3698, 0x3699, 0x0000, 0x3694, 0x3695, 0x3696, 0x0000, 0x3691,
/*A0*/	0x3692, 0x3693, 0x3690, 0x369a, 0x372f, 0x3009, 0x370d, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
/*B0*/	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
/*C0*/	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
/*D0*/	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
/*E0*/	0x0700, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
/*F0*/	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
};

int get_key()
{
static int e_code_bytes = 0;	/* For E0 or E1 byte count */
static int big_scan = 0;	/* Build up a composite scan code */
#ifndef KEYTEST
static int pending_keyscan = 0;	/* A scancode for which we expect a bios key */
#endif

static int current_flags = 0;
static int last_flags = 0;
static int forced_flags[] = {0, KEY_SHIFT, KEY_CTRL, KEY_ALT};

   int ch, scan, key_flags;
   int scand;
   int i, key_ready;

#ifdef KEYTEST
   last_stuffed  = last_bios_key = last_ctrl_flg = -1;
#endif

   /* Make sure we have an AT keyboard */
   if( kbd_type < 0 )
#ifdef __BORLANDC__
   {
      kbd_type = 0;

      for(i=0; i<16;i++)
      {
         _AH = 0x01; geninterrupt(0x16); if((_FLAGS&0x40)) break;
         _AH = 0x00; geninterrupt(0x16);
      }
      _AH = 0x05; _CX= 0xFFFF; geninterrupt(0x16);
      for(i=0; i<16;i++)
      {
         _AH = 0x11; geninterrupt(0x16); if((_FLAGS&0x40)) break;
         _AH = 0x10; geninterrupt(0x16);
	 if( _AX == 0xFFFF )
	 {
	     kbd_type = 0x10;
	     break;
	 }
      }
   }
#else
      kbd_type = 0;
#endif

#ifdef KEYTEST
   if( bios_stuff ) { ch = bios_stuff; bios_stuff = key_ready = 0; } else {
#endif
#ifdef __BORLANDC__
   _AH = kbd_type+1; geninterrupt(0x16);
   ch = _AX;
   key_ready = !!(_FLAGS&0x40);

   /* Collect non-bios keys */
   if( key_ready == 0 && (ch&0xFF00) == 0xFF00 && pending_keyscan )
      key_ready = 1;
   if( key_ready == 1 && pending_keyscan )
   {
      ch = (pending_keyscan<<8) + ((pending_keyscan&0xFF00)?0xE0:0);
      key_ready = 2;
   }

#else
   key_ready = !_bios_keybrd(_KEYBRD_READY);
#endif

   if(key_ready == 0)
   {
#ifdef __BORLANDC__
      _AH = kbd_type; geninterrupt(0x16);
      ch = (unsigned) _AX;
#else
      ch = _bios_keybrd(_KEYBRD_READ);
#endif
   }
#ifdef KEYTEST
   }
#endif

   /* -- Add fake scan codes for shift changes on XT keyboards -- */
   if( kbd_type == 0 && key_ready == 1 )
   {
static int last_flgs = 0;
      int flgs = (_bios_keybrd(_KEYBRD_SHIFTSTATUS) & 0xF);
      if( flgs != last_flgs )
      {
	 key_ready = 0;
         ch = 0xFF00;
	 if( ( flgs^last_flgs ) & 1 )
	    ch = 0xFF36 + ((flgs&1)?0:0x80), last_flgs ^= 1;
	 else if( ( flgs^last_flgs ) & 2 )
	    ch = 0xFF2A + ((flgs&2)?0:0x80), last_flgs ^= 2;
	 else if( ( flgs^last_flgs ) & 4 )
	    ch = 0xFF1D + ((flgs&4)?0:0x80), last_flgs ^= 4;
	 else if( ( flgs^last_flgs ) & 8 )
	    ch = 0xFF38 + ((flgs&8)?0:0x80), last_flgs ^= 8;
      }
   }

   if( key_ready == 1 )
   {
      /* Generate the codes for double shift key flagging */
      if( last_flags != current_flags )
      {
         for(i=((current_flags|last_flags)&0x7FFF); (i&1) == 0; i>>=1)
            ;
         if( i & -2 )
	 {
            if( current_flags > last_flags )
            {
               last_flags = current_flags;
	       i = current_flags;
	       if( i & KEY_CTRL2 ) i |= KEY_CTRL;
	       if( i & KEY_SHIFT2 ) i |= KEY_SHIFT;
	       return KEY_SPECIAL + (i & 0x7F00);
            }
	    last_flags = 0;
            return KEY_SPECIAL;
         }
      }
      return -1;
   }

   /* Is this a scan code not a bios key ? */
   if( ( ch & 0xFF00 ) == 0xFF00 ) 
   {
      ch &= 0xFF;

      if( e_code_bytes ) { e_code_bytes--; big_scan = (big_scan << 8) + ch; }
      else if( ch == 0xE0 ) { e_code_bytes=1; big_scan = 0xE0; }
      else if( ch == 0xE1 ) { e_code_bytes=2; big_scan = 0; }
      else
         big_scan = ch;

      if( !e_code_bytes )
      {
static char shift_scans[8] = { 0x1D, 0x2A, 0x36, 0x38,/*0x3A, 0x45, 0x46,*/0 };
static int  shift_flags[8] = { KEY_CTRL, KEY_SHIFT, KEY_SHIFT2, KEY_ALT };
         char * p;
	 int sw_flag = 0;
         if( (p=strchr(shift_scans, (ch&0x7F))) != 0 )
	 {
	    switch(big_scan&0x7F7F)
	    {
	    case 0x601D: sw_flag = KEY_CTRL2; break;
	    case 0x602A: /* Un-numlock: The BIOS actually does this! */ break;
            case 0x1D45: /* Pause */ break;
	    case 0x6038: sw_flag = KEY_ALTGR; break;
	    default:     sw_flag = shift_flags[p-shift_scans]; break;
	    }
	    if( ch & 0x80 )
	       current_flags &= ~sw_flag;
	    else
	       current_flags |= sw_flag;
	 }
	 else if( ( ch & 0x80 ) == 0 && STATE_ALTCD == 0 )
	    pending_keyscan = big_scan;
      }
/*
cprintf("\rec %d, bs=%04x, fl=%04x, sc=%02x\r\n", e_code_bytes, big_scan, current_flags, ch);
*/
      return KEY_SPECIAL + KEY_DIRECT + KEY_OTHER + ch;
   }

#ifdef KEYTEST
   if( key_ready == 2 )
      last_stuffed  = ch;
   else
      last_bios_key = ch;
   last_ctrl_flg = current_flags;
#endif

   pending_keyscan = 0;	/* We've got something like a bios key */

   if( ch == 0 ) return KEY_OTHER + 0x03;

   scan = ((ch>>8)&0xFF);
   ch &= 0xFF;

   if( scan == 0 ) return ch + KEY_DIRECT;
   if( scan == 0xE0 ) scan = 0xA4 + 2*(ch < ' ');

   key_flags = 0;
   if( ch == 0xE0 || (key_ready != 2 && scan >= 0x97 && scan <= 0xA3 ))
   { key_flags |= KEY_PAD; ch = 0; }

   /* Check for forced flags */
   if( ch == 0 && current_flags == 0 && key_ready != 2 )
      key_flags |= forced_flags[(bioskeydecode[scan]>>12)&3];

   if( current_flags & (KEY_CTRL|KEY_CTRL2) )   key_flags |= KEY_CTRL;
   if( current_flags & (KEY_SHIFT|KEY_SHIFT2) ) key_flags |= KEY_SHIFT;
   key_flags |= (current_flags & (KEY_ALT|KEY_ALTGR));

   if( ch && key_ready != 2 ) switch(key_flags)
   {
   case KEY_CTRL: if( (ch >= ' ' && ch < '@') || ch > 127) break;
   case 0:
   case KEY_SHIFT:
      switch((bioskeydecode[scan]>>8)&7)
      {
      case 1:
         return ch;
      case 2:
         if( key_flags == 0 ) return ch + KEY_OTHER;
         break;
      case 3:
         if( (key_flags & ~KEY_CTRL) == 0 ) return ch + KEY_OTHER;
         break;
      case 4:
         if( (key_flags & ~KEY_SHIFT) == 0 ) return ch;
         if( key_flags == KEY_CTRL ) return (ch&0x1F);
         break;
      case 5:
         if( key_flags == 0 ) return ch;
         break;
      case 6:
         if( key_flags == 0 ) return ch + KEY_PAD;
         break;
      case 7: return ch + KEY_PAD;
      }
   }

   key_flags |= KEY_SPECIAL;

   if( ((bioskeydecode[scan]>>8)&6) == 6 && kbd_type )
      key_flags ^= KEY_PAD;
   
   if( ((bioskeydecode[scan]>>8)&7) == 4 &&
       ( ch || ( key_ready == 2 && ((bioskeydecode[scan]>>12)&3) != 0 ) ) )
      ;
   else if((bioskeydecode[scan]&0xFF) &&
           (!(key_flags&KEY_PAD) || (bioskeydecode[scan]&0x700)>=0x600))
      return key_flags + (bioskeydecode[scan]&0xFF);

   /* Now the last few specials ... */
   if( key_flags & KEY_PAD )
   {
      /* Keypad-'*' where it conflicts with PrintScrn */
      if( scan == 0x37 ) return key_flags + '*';

      /* Num lock */
      if( scan == 0x45 ) return 0x9B + (key_flags & ~KEY_PAD);

      /* Pause/HoldScrn */
      if( scan == 0x46 ) return 0x9C + (key_flags & ~KEY_PAD);
   }
   else
   { 
      /* Pause/HoldScrn again */
      if( scan == 0x45 ) return 0x9C + key_flags;

      /* Scroll lock */
      if( scan == 0x46 ) return 0x9D + key_flags;

      /* Caps lock */
      if( scan == 0x3A ) return 0x9F + key_flags;

      /* Key 56h, |\ or <> */
      if( scan == 0x56 ) return '<' + key_flags;

      /* Print screen */
      if( scan == 0x37 ) return 0x9E + key_flags;

      /* Print screen */
      if( scan == 0x72 )
         return 0x9E + key_flags | (current_flags==0?KEY_CTRL:0);

      /* SysRq */
      if( scan == 0x54 )
         return 0x9E + key_flags | (current_flags==0?KEY_ALT:0);
   }

   return key_flags + KEY_DIRECT + scan;
}

